"use strict";
//
// This is an example of running tasks on a single core.
//
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var _this = this;
Object.defineProperty(exports, "__esModule", { value: true });
var task_1 = require("../lib/task");
var single_core_scheduler_1 = require("../lib/single-core-scheduler");
var console_1 = require("console");
var taskA = task_1.Task.register(// All tasks must be registered globally so they exist in both the master and worker processes.
"TaskA", // Tasks must be named so they can be identified in the work proceses.
function () { return __awaiter(_this, void 0, void 0, function () {
    return __generator(this, function (_a) {
        console.log("!!!! Task A");
        return [2 /*return*/, 1]; // <-- This could be a complex time consuming calculation. Results must be serializable to JSON.
    });
}); });
var taskB = task_1.Task.register("TaskB", function () { return __awaiter(_this, void 0, void 0, function () {
    return __generator(this, function (_a) {
        console.log("!!!! Task B");
        return [2 /*return*/, 2]; // <-- Any of these task functions can be asynchronous and return a promise.
    });
}); });
var taskC = task_1.Task.register("TaskC", function (a, b) { return __awaiter(_this, void 0, void 0, function () {
    return __generator(this, function (_a) {
        console.log("!!!! Task C");
        return [2 /*return*/, a + b]; // <-- This task is dependent on pre-computed inputs from other tasks.
    });
}); });
//
// Your function to wire together a task with dependencies.
//
function createTask() {
    //
    // Your code here decided how to build the task graph.
    // Here we are saying that taskA and taskB are inputs to task C.
    //
    return taskC.create(taskA.create(), taskB.create());
}
//
// Run application code.
//
function main(scheduler) {
    return __awaiter(this, void 0, void 0, function () {
        var task, answer;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    task = createTask();
                    return [4 /*yield*/, task.run(scheduler)];
                case 1:
                    answer = _a.sent();
                    console.log("The answer: " + answer);
                    console_1.assert(answer === 3);
                    return [2 /*return*/];
            }
        });
    });
}
//
// Bootstrap the cluster.
//
function bootstrap() {
    return __awaiter(this, void 0, void 0, function () {
        var scheduler;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    scheduler = new single_core_scheduler_1.SingleCoreScheduler();
                    return [4 /*yield*/, main(scheduler)];
                case 1:
                    _a.sent();
                    return [2 /*return*/];
            }
        });
    });
}
bootstrap()
    .catch(function (err) { return console.error(err && err.stack || err); });
//# sourceMappingURL=single-core.js.map