"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var uuid_1 = require("uuid");
//
// Defines a parallelizable asynchronous task.
//
var TaskDef = /** @class */ (function () {
    function TaskDef(taskName, taskFn) {
        this.taskName = taskName;
        this.taskFn = taskFn;
    }
    //
    // Retreive the user-defined name of the task.
    //
    TaskDef.prototype.getTaskName = function () {
        return this.taskName;
    };
    //
    // Get the function that is invoked to perform the task.
    //
    TaskDef.prototype.getTaskFn = function () {
        return this.taskFn;
    };
    //
    // Instantiate a task from the definition.
    //
    TaskDef.prototype.create = function () {
        var inputs = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            inputs[_i] = arguments[_i];
        }
        var wrappedTasks = inputs.map(this.wrapTask); // Wrap direct inputs as tasks.
        return new Task(wrappedTasks, this);
    };
    //
    // Wrap direct values in tasks where necessary.
    //
    TaskDef.prototype.wrapTask = function (input) {
        if (typeof (input.run) === "function") {
            // It's a task.
            return input;
        }
        else {
            // It's a direct value, so wrap it in a task.
            return new WrappedValue(input);
        }
    };
    return TaskDef;
}());
//
// Interface to a parallelizable asynchronous task.
//
var Task = /** @class */ (function () {
    function Task(inputTasks, taskDef) {
        this.taskId = uuid_1.v4();
        this.inputTasks = inputTasks;
        this.taskDef = taskDef;
    }
    //
    // Register a task.
    //
    Task.register = function (taskName, taskFn) {
        var taskDef = new TaskDef(taskName, taskFn);
        Task.taskDefs[taskDef.getTaskName()] = taskDef;
        return taskDef;
    };
    // 
    // Look up a task.
    //
    Task.lookup = function (taskDefId) {
        return Task.taskDefs[taskDefId];
    };
    //
    // Run multiple tasks in parallel and collect their results as an array.
    //
    Task.all = function (inputTasks) {
        return new AllTask(inputTasks);
    };
    //
    // Retreive the unique ID for the task.
    //
    Task.prototype.getTaskId = function () {
        return this.taskId;
    };
    //
    // Retreive the task definition.
    //
    Task.prototype.getTaskDef = function () {
        return this.taskDef;
    };
    //
    // Run a task and all it's input tasks as soon as workers are available.
    // Returns a promise that is resolved when the task and its inputs have completed successfully.
    // Rejects the promise if the task or any input throws an error.
    //
    Task.prototype.run = function (scheduler) {
        return __awaiter(this, void 0, void 0, function () {
            var inputs;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, Promise.all(this.inputTasks.map(function (inputTask) { return inputTask.run(scheduler); }))];
                    case 1:
                        inputs = _a.sent();
                        return [4 /*yield*/, scheduler.runTask(inputs, this)];
                    case 2: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    //
    // Lookup table for tasks.
    // Allows task to be found by name when request to run in a worker process.
    //
    Task.taskDefs = {};
    return Task;
}());
exports.Task = Task;
//
// A task that wraps up a direct value.
//
var WrappedValue = /** @class */ (function () {
    function WrappedValue(value) {
        this.value = value;
    }
    //
    // Retreive the unique ID for the task.
    //
    WrappedValue.prototype.getTaskId = function () {
        return "wrapped"; // This doesn't need a distinct ID because it's not to be transmitted to a worker process.
    };
    //
    // Retreive the task definition.
    //
    WrappedValue.prototype.getTaskDef = function () {
        var _this = this;
        return new TaskDef("wrapped", function () { return __awaiter(_this, void 0, void 0, function () { return __generator(this, function (_a) {
            return [2 /*return*/];
        }); }); });
    };
    //
    // Simply returns the wrapped value.
    //
    WrappedValue.prototype.run = function (scheduler) {
        return this.value;
    };
    return WrappedValue;
}());
//
// A task that aggregates a list of tasks.
//
var AllTask = /** @class */ (function () {
    function AllTask(inputTasks) {
        this.inputTasks = inputTasks;
    }
    //
    // Retreive the unique ID for the task.
    //
    AllTask.prototype.getTaskId = function () {
        return "all"; // This doesn't need a distinct ID because it's not to be transmitted to a worker process.
    };
    //
    // Retreive the task definition.
    //
    AllTask.prototype.getTaskDef = function () {
        var _this = this;
        return new TaskDef("all", function () { return __awaiter(_this, void 0, void 0, function () { return __generator(this, function (_a) {
            return [2 /*return*/];
        }); }); }); //TODO: Is there a better design that doesn't need these fake implementations?
    };
    //
    // Simply returns the wrapped value.
    //
    AllTask.prototype.run = function (scheduler) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, Promise.all(this.inputTasks.map(function (inputTask) { return inputTask.run(scheduler); }))];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    return AllTask;
}());
//# sourceMappingURL=task.js.map